## libraries etc ####
library(FITfileR) # for reading and processing .fit files
library(ggplot2) # for data processing and plotting
library(dplyr) # for data processing and plotting
library(scales) # for date format in a plot
library(lubridate) # for transforming min:sec to sec and vice versa
library(xts) # working with Time Series for smoothing grade and pace (outlier removal and replacement)
library(forecast) # for interpolating the Time Series during outlier replacement
source("functions/tsclean2.R")
source("functions/tsoutliers2.R")

plotTheme = 
  theme_bw(base_size=8) +
  theme(axis.text=element_text(size=8,
                               colour="black"))

## reading data from a .fit file ####

# reading the .fit file and the records of an activity
fitFile = FITfileR::readFitFile(
  "input/xortiathsTrail22km.fit")
fitFileRec = FITfileR::records(fitFile)

# some .fit files have multiple records, others have a single record in a tibble after reading it before
singleTibble = TRUE
if(is.null(dim(fitFileRec))) # dimensions is NULL if it's a single tibble
{
  singleTibble = FALSE
}

# in case you want to check the data attributes, units etc
testDataSfileSession = FITfileR::getMessagesByType(
  fitFile, 
  "session"
  )
# for example,
testDataSfileSession$start_position_lat # coordinates usually in degrees (probably in WGS84)
testDataSfileSession$total_distance # distance in meters
testDataSfileSession$avg_speed # speed in m/s
testDataSfileSession$total_ascent # altitude in meters
testDataSfileSession$max_cadence # cadence is in rpm, i.e. rotation of a leg; you hence need to multiply it by two to obtain steps per minute

# first, you need to bind the parts of the fit file, if it's not single-tibbled
names(fitFileRec)
names(fitFileRec$record_1) # for GAP, keep only specific columns. Unfortunately, different .fit files have different names of the same thing, and hence you need to provide the names manually, after looking them up here
if(!singleTibble)
{
  data = fitFileRec %>% 
    bind_rows() %>% 
    arrange(timestamp) %>% # arrange row order by the time of measurement
    select( # provide the names manually here
      c(timestamp, position_lat, position_long, distance, speed, altitude))
} else 
{
  data = fitFileRec %>%
    select( # or here, in case of a single tibble
      c(timestamp, position_lat, position_long, distance, speed, altitude))
}

# some initial checks
names(data)
nrow(data)
summary(data) # checking for outliers (some .fit files have crazy values, which will be taken care of soon)

dataF = as.data.frame(data) # i prefer working with data frames
names(dataF) = c("timestamp", "lat", "lon", "distance", "speed", "altitude") # same for any .fit file from hereon

## some more checks to refine the data ####
# are the timestamps unique?
length(unique(dataF$timestamp))  == length(dataF$timestamp)

# is any distance or timestamp lower or equal than its previous?
dataF$distanceSmaller = 0 # new column with all 0's, i.e. all previous distances are lower
dataF$timestampSmaller = 0 # new column with all 0's, i.e. all previous timestamps are earlier
for(i in 2:nrow(dataF)) # bad way, lazy day...
{
  distanceRef = dataF$distance[i]
  distanceCheck = dataF$distance[1:(i-1)]
  dataF$distanceSmaller[i] = mean(distanceRef <= distanceCheck)
  
  timestampRef = dataF$timestamp[i]
  timestampCheck = dataF$timestamp[1:(i-1)]
  dataF$timestampSmaller[i] = mean(timestampRef <= timestampCheck)
}
# which are not OK
which(dataF$distanceSmaller > 0)
which(dataF$timestampSmaller > 0)

# this is another variable we'll need for GAP, the distance moved between consecutive measurements
dataF$distDiff = dataF$distance[1]
for(i in 2:nrow(dataF))
{
  dataF$distDiff[i] = dataF$distance[i] - dataF$distance[i-1]
}
# hist(dataF$distDiff)
which(dataF$distDiff <= 0) == which(dataF$distanceSmaller > 0) # should give similar results to the previous

# will finally remove such observations, plus outliers of latitude, longitude, speed and altitude
dataF = dataF[which(
  dataF$distanceSmaller <= 0 &
  dataF$timestampSmaller <= 0 &
  dataF$lat <= quantile(dataF$lat)[4] + 1.5 * IQR(dataF$lat) &
  dataF$lat >= quantile(dataF$lat)[2] - 1.5 * IQR(dataF$lat) &
  dataF$lon <= quantile(dataF$lon)[4] + 1.5 * IQR(dataF$lon) &
  dataF$lon >= quantile(dataF$lon)[2] - 1.5 * IQR(dataF$lon) &
  dataF$speed <= quantile(dataF$speed)[4] + 1.5 * IQR(dataF$speed) &
  dataF$speed >= quantile(dataF$speed)[2] - 1.5 * IQR(dataF$speed) &
  dataF$altitude <= quantile(dataF$altitude)[4] + 1.5 * IQR(dataF$altitude) &
  dataF$altitude >= quantile(dataF$altitude)[2] - 1.5 * IQR(dataF$altitude)
), ]
summary(dataF)

# adding the pace in seconds
dataF$paceSec = 1 / (dataF$speed / 1000)

# checking the pace
durSec = seq(3540, 0, -120)
durTimeSec = seconds_to_period(durSec)
durTimeMinSec = sprintf(
  '%2d:%02d', minute(durTimeSec), second(durTimeSec)
  )
# plotting with the timestamp
ggplot(dataF, 
       aes(timestamp, 
           paceSec)) +
  geom_line() +
  scale_y_reverse(breaks=durSec,
                  labels=durTimeMinSec) +
  scale_x_datetime(labels = scales::date_format("%H:%M:%S")) +
  labs(x="Time (UTC)",
       y="Distance (km)") +
  plotTheme
# plotting with the distance
ggplot(dataF, 
       aes(x=distance/1000, 
           y=paceSec)) +
  geom_line() +
  scale_y_reverse(breaks=durSec,
                  labels=durTimeMinSec) +
  scale_x_continuous(breaks=seq(0, 100, 2)) +
  labs(x="Distance (km)",
       y="Pace ( / km)") +
  plotTheme

# elevation profile
ggplot(dataF, 
       aes(x=distance/1000, 
           y=altitude)) +
  geom_line() +
  scale_x_continuous(breaks=seq(0, 100, 2)) +
  labs(x="Distance (km)",
       y="Elevation (m)") +
  plotTheme

# for grade, I need the altitudinal difference
dataF$altDiff = 0
for(i in 2:nrow(dataF))
{
  dataF$altDiff[i] = dataF$altitude[i] - dataF$altitude[i-1]
}
dataF$gradePercent = dataF$altDiff / dataF$distDiff * 100

# checking grade alongside elevation
ggplot(dataF, 
       aes(x=distance/1000, 
           y=altitude)) +
  geom_line() +
  geom_line(data=dataF, 
            aes(x=distance/1000, 
                y=gradePercent)) +
  scale_x_continuous(breaks=seq(0, 100, 2)) +
  labs(x="Distance (km)",
       y="Elevation (m)") +
  plotTheme
summary(dataF$gradePercent)
# there might be some outliers, which I will not touch for now?

# checking grade alone
ggplot(dataF, 
       aes(x=distance/1000, 
           y=gradePercent)) +
  geom_line() +
  scale_x_continuous(breaks=seq(0, 100, 2)) +
  labs(x="Distance (km)",
       y="Grade (%)") +
  plotTheme

# smoothing out the grade
xtsSecGrade = data.frame(
  timestamp=dataF$timestamp,
  gradePercent=dataF$gradePercent)
#
gradePercentTs = xts::xts(
  xtsSecGrade$gradePercent,
  order.by=xtsSecGrade$timestamp)
gradePercentTsClean = tsclean2(gradePercentTs,
                               lowFenceDistIQR=1.5,
                               highFenceDistIQR=1.5)


dataF$gradePercentClean = as.vector(gradePercentTsClean)

# checking grade alongside elevation
ggplot(dataF, 
       aes(x=distance/1000, 
           y=altitude)) +
  geom_line() +
  geom_line(data=dataF, 
            aes(x=distance/1000, 
                y=gradePercentClean)) +
  scale_x_continuous(breaks=seq(0, 100, 2)) +
  scale_y_continuous(breaks=seq(-500, 2000, 20)) +
  labs(x="Distance (km)",
       y="Elevation (m)") +
  plotTheme

# checking grade alone
ggplot(dataF, 
       aes(x=distance/1000, 
           y=gradePercentClean)) +
  geom_line() +
  scale_x_continuous(breaks=seq(0, 100, 2)) +
  scale_y_continuous(breaks=seq(-500, 2000, 10)) +
  labs(x="Distance (km)",
       y="Grade (%)") +
  plotTheme

# smoothing out the pace
xtsSecPace = data.frame(
  timestamp=dataF$timestamp,
  paceSec=dataF$paceSec)
#
paceSecTs = xts::xts(
  xtsSecPace$paceSec,
  order.by=xtsSecPace$timestamp)
paceSecTsClean = tsclean2(paceSecTs,
                               lowFenceDistIQR=1.5,
                               highFenceDistIQR=1.5)


dataF$paceSecTsClean = as.vector(paceSecTsClean)


# checking
durSec = seq(3540, 0, -120)
durTimeSec = seconds_to_period(durSec)
durTimeMinSec = sprintf(
  '%2d:%02d', minute(durTimeSec), second(durTimeSec)
)
# checking with the distance
ggplot(dataF, 
       aes(x=distance/1000, 
           y=paceSecTsClean)) +
  geom_line() +
  scale_y_reverse(breaks=durSec,
                  labels=durTimeMinSec) +
  scale_x_continuous(breaks=seq(0, 100, 2)) +
  labs(x="Distance (km)",
       y="Pace ( / km)") +
  plotTheme


## flat pace (GAP) from actual hill pace ####
dataFin = dataF # will hereforth work with this one

dataFin[1, "distDiff"] = dataFin[1, "distance"]

dataFin$flatCoef = 1 + 0.02869556*dataFin$gradePercentClean + 0.001520768*dataFin$gradePercentClean^2 # that's the quadratic calculation of the coefficient for the GAP point estimate based on the 

dataFin$flatPaceSec = dataFin$paceSecTsClean / dataFin$flatCoef # that's the GAP point estimate

# checking
durSec = seq(3540, 0, -120)
durTimeSec = seconds_to_period(durSec)
durTimeMinSec = sprintf(
  '%2d:%02d', minute(durTimeSec), second(durTimeSec)
)
# checking with the distance as STRAVA
ggplot(dataFin, 
       aes(x=distance/1000, 
           y=flatPaceSec)) +
  geom_line() +
  scale_y_reverse(breaks=durSec,
                  labels=durTimeMinSec) +
  scale_x_continuous(breaks=seq(0, 100, 2)) +
  labs(x="Distance (km)",
       y="GAP (min/km)") +
  plotTheme

##  estimate average flat pace (GAP)
GAPmeanSec = weighted.mean(
  dataFin$flatPaceSec, 
  w=dataFin$distDiff # I'm weighting the GAP point estimates by the distance moved until the next measurement point
)
lubridate::seconds_to_period(GAPmeanSec)

## average hill pace from target flat pace (e.g. threshold pace) ####

# the idea here is to estimate what the actual pace will be, given a pace on flat. In that way, you can additionally estimate the time for completing the course, and the average actual (hill) pace.

dataFin = dataF

dataFin[1, "distDiff"] = dataFin[1, "distance"]

dataFin$targetFlatPace = 250 # seconds

dataFin$flatCoef = 1 + 0.02869556*dataFin$gradePercentClean + 0.001520768*dataFin$gradePercentClean^2

dataFin$targetHillPace = dataFin$flatCoef * dataFin$targetFlatPace

dataFin$targetHillSpeed = 1000 / dataFin$targetHillPace

dataFin$targetHillTimeDiff = dataFin$distDiff / dataFin$targetHillSpeed

# checking
durSec = seq(3540, 0, -60)
durTimeSec = seconds_to_period(durSec)
durTimeMinSec = sprintf(
  '%2d:%02d', minute(durTimeSec), second(durTimeSec)
)
# with the distance
ggplot(dataFin, 
       aes(x=distance/1000, 
           y=targetHillPace)) +
  geom_line() +
  scale_y_reverse(breaks=durSec,
                  labels=durTimeMinSec) +
  geom_hline(yintercept=unique(dataFin$targetFlatPace)) +
  labs(x="Distance (km)",
       y="Pace ( / km)") +
  plotTheme

##  estimated pace and time for the course
# time
targetHillHoursTot = (sum(dataFin$targetHillTimeDiff) / 60) %/% 60 # hours
targetHillMinsTot = (sum(dataFin$targetHillTimeDiff) / 60) %% 60 # minutes
print(
  paste0(
    "Estimated time: ",
    targetHillHoursTot, " hour(s) and ",
    round(targetHillMinsTot), " minute(s)."))

# pace
targetHillSpeedTot = sum(dataFin$distDiff) / sum(dataFin$targetHillTimeDiff)
targetHillPaceTot = 1000 / targetHillSpeedTot
targetHillPaceMinsTot = targetHillPaceTot %/% 60
targetHillPaceSecsTot = targetHillPaceTot %% 60
print(
  paste0(
    "Estimated pace: ",
    targetHillPaceMinsTot, ":",
    round(targetHillPaceSecsTot), " min/km."))

## segmenting a course to grade or GAP regions ####

# this is an attempt to mimic Garmin's PacePro feature when you ask it for pace per elevation change (not for every km or mile)

library(changepoint)

segCount = 31 # let's say you want to cluster/split the course to that many segments of spatially more homogeneous pace

gradeGroup = changepoint::cpt.mean(
  data=dataFin$targetHillPace,
  penalty="AIC",
  pen.value=0,
  method="BinSeg",
  Q=segCount-1,
  test.stat="Normal",
  class=TRUE,
  param.estimates=TRUE,
  minseglen=30
  )
plot(gradeGroup)

# refining the change points and the segments
splitPointCounts = gradeGroup@cpts - c(1, 
                    gradeGroup@cpts[1:(length(gradeGroup@cpts)-1)])
splitPointCounts[length(splitPointCounts)] = splitPointCounts[length(splitPointCounts)] + 1

# putting it on the data frame
dataFin$split = rep(
  1:(length(splitPointCounts)),
  splitPointCounts
)
dataFin$splitHillPace = rep(
  gradeGroup@param.est$mean,
  splitPointCounts
)

# checking
durSec = seq(3540, 0, -60)
durTimeSec = seconds_to_period(durSec)
durTimeMinSec = sprintf(
  '%2d:%02d', minute(durTimeSec), second(durTimeSec)
)
# with the distance 
ggplot(dataFin, 
       aes(x=distance/1000, 
           y=splitHillPace)) +
  geom_line() +
  scale_y_reverse(breaks=durSec,
                  labels=durTimeMinSec) +
  scale_x_continuous(breaks=seq(0, 100, 2)) +
  geom_hline(
    yintercept=unique(dataFin$targetFlatPace),
    linetype="dotted") + # the target flat pace for reference with a dotted line
  geom_hline(
    yintercept=targetHillPaceTot,
    linetype="dashed") + # the mean hill pace for reference with a dashed line
  labs(x="Distance (km)",
       y="Pace ( / km)") +
  plotTheme

splitPaceSec = unique(dataFin$splitHillPace)
splitDist = aggregate(
  distDiff ~ split, 
  data=dataFin, 
  FUN=sum)
splitGrade = aggregate(
  gradePercentClean ~ split, 
  data=dataFin, 
  FUN=mean)
# putting all results in a data frame
splitDF = data.frame(
  split=1:segCount,
  distKm=round(
   splitDist$distDiff/1000,
   2),
  meanGrade=round(
   splitGrade$gradePercentClean,
   2),
  pace=round(
   lubridate::seconds_to_period(splitPaceSec))
           )
splitDF

## distribution of course's grade #####

# this can also be important summary of a course (not only the total ascent)...

ggplot(
  dataF, 
  aes(
    x=gradePercentClean
    )
  ) + 
  geom_histogram(
    aes(
      weight=distDiff/1000
        )
  ) +
  labs(
    x="Grade (%)",
    y="Length (km)"
  ) +
  plotTheme
